import argparse
import json
import os
import prompter as p
import sys

def dump_files(data:list, chunks:list, failed_files:list):

    # Convert training_data paths into strings for serialization.
    data_str = list()
    for d in data: 
        data_str.append(str(d))
    # Save each used list as a file for other operations.
    with open('training_data.json', 'w') as f:
        json.dump(data_str, f)
    with open('chunks.json', 'w') as f:
        json.dump(chunks, f)
    with open('failed_files.json', 'w') as f:
        json.dump(failed_files, f)

if __name__ == "__main__":
    # Check that environment variables are set up.
    if "OPENAI_API_KEY" not in os.environ:
        print("You must set an OPENAI_API_KEY using the Secrets tool", file=sys.stderr)
    # Create positional and optional command arguments
    parser = argparse.ArgumentParser(description='Create a vector store to pass contextual data to an LLM prompt')
    parser.add_argument('data_path', help='the location of the training data', default=None)
    parser.add_argument('file_extension', help='the file extension to be searched for and then parsed, e.g., *.java', default=None)
    parser.add_argument('-d', '--dump-files', action='store_true', dest='dump_key_files', help='dump key info into files', default=False)
    parser.add_argument('-f', '--parse-files', action='store', dest='parse_files', help='one or more files that contain prompt questions', nargs='*')
    parser.add_argument('-p', '--master-prompt', action='store', dest='master_prompt', help='master prompt path, default is `training/master.prompt`', default='training/master.prompt')
    parser.add_argument('-c', '--show-context', action='store_true', dest='show_context', help='show the context passed to the LLM', default=False)
    parser.add_argument('-s', '--show-statistics', action='store_true', dest='show_statistics', help='show the parsing and chunking statistics', default=False)
    args = parser.parse_args()

    data, chunks, failed_files = p.chunk(args.data_path, args.file_extension)
    if args.show_statistics:
        p.print_parsing_statistics(data, failed_files)
        p.print_token_statistics(chunks)
    if args.dump_key_files:
        dump_key_files(data, chunks, failes_files)
    store = p.train(chunks)
    if args.parse_files == None:
        p.prompter(store, args.master_prompt, args.show_context)
    else:
        p.files_prompter(store, args.master_prompt, args.show_context)


import json
from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts import Prompt
from langchain.embeddings import OpenAIEmbeddings
from langchain.vectorstores import Chroma
import os
from data_chunker import parser as JCParser
from data_chunker import java_code as JCChunker


# Check that environment variables are set up.
if "OPENAI_API_KEY" not in os.environ:
    raise ValueError("You must set an OPENAI_API_KEY environment variable value")

def chunker(directory, file_extension:str="*.java", write_to_disk:bool=True, outdir:str="."):
    """Leverages data-chunker package to parse & chunk text-based code files into LLM-consumable tokens. Currently only supports java (*.java)
    
    Defaults to writing chunks to disk; you may write to memory by using write_to_disk=False, but be sure to store the chunks
    as a variable for later consumption.

    Returns chunked tokens that can be added to a vector store.
    """
    
    if file_extension != "*.java":
        raise ValueError("This file type is not supported yet.")

    training_data = list()
    
    training_data = JCParser.get_file_list(directory, file_extension=file_extension)
    # Chunk data using the files in the training data
    chunks = []
    failed_files = []
    for file in training_data:
        codelines = JCParser.get_code_lines(file)
        try:
            tree = JCChunker.parse_code(file, codelines)
        except JCChunker.ParseError as e:
            failed_files.append(str(file) + ": " + str(e))
        if tree != None:
            try:
                chunks = chunks + JCChunker.chunk_constants(tree)
                chunks = chunks + JCChunker.chunk_constructors(tree, codelines)
                chunks = chunks + JCChunker.chunk_fields(tree, codelines)
                chunks = chunks + JCChunker.chunk_methods(tree, codelines)
            except JCChunker.ChunkingError as e:
                failed_files.append(str(file) + ": " + str(e))
        else:
            failed_files.append(str(file) + ", has no tree!")
    
    if(write_to_disk):
        if(outdir==None):
            raise ValueError("You must specify a directory (outdir) to write chunks to disk.")
        training_data_str = list()
        for data in training_data: 
            training_data_str.append(str(data))
        # Save each used list as a file for other operations.
        with open(f"{outdir}/training_data.json", 'w') as f:
            json.dump(training_data_str, f)
        with open(f"{outdir}/chunks.json", 'w') as f:
            json.dump(chunks, f)
        with open(f"{outdir}/failed_files.json", 'w') as f:
            json.dump(failed_files, f)

    return chunks


def create_store(data:list[str]=None, name:str="langchain_store", persistdir:str="db", model:str="text-embedding-ada-002", filepath:str="chunks.json"):
    """Create the vector store to utilize for other commands (currently limited to OpenAI). Consumes a list of JSON format data strings
    to create a local vector store and then utilizes an API call to OpenAI embedders.
    
    Defaults to using the text-embedding-ada-002 model.
    
    Returns a Chroma store."""
    
    #TODO: make an ephemeral option
    #TODO: make a for loop to iterate over a potential suite of JSON files

    str_chunks = []

    # Only use the data string if we explicitly provide one (default behavior is to use on-disk chunks)
    if(data==None):
        with open(filepath, 'r') as f:
            chunks = json.load(f)

        for chunk in chunks:
            str_chunks.append(str(chunk))

    else:
        for chunk in data:
            str_chunks.append(str(chunk))

    store = Chroma(collection_name=name,
                embedding_function=OpenAIEmbeddings(model=model),
                persist_directory=persistdir).from_texts(str_chunks, OpenAIEmbeddings(model=model))
        
    return store
    
def search_store(store: Chroma, text: str):
    """Perform a similarity search against the Chroma vector store based on the text provided."""
    store_chunks = store.similarity_search(text)

    return store_chunks

def prompt(store: Chroma, show_context=False, templateFilePath:str=None, model:str="gpt-3.5-turbo", write_to_disk=True):
    """Setup a chat session with the LLM (currently limited to OpenAI). The session maintains history by storing the
    previous answers into a history list and appending them to each future prompt, meaning there is a limit for number of 
    questions per individual session (you will eventually reach the token limit per model).

    Defaults to using OpenAI's GPT-3.5-Turbo
    
    Will continue the chat session until the user types 'exit' as their prompt."""

    history = ""

    # Load the promptTemplate for model context :
    if templateFilePath != None:
        with open(templateFilePath, "r") as f:
            promptTemplate = f.read()
    else:
        promptTemplate = """You are a world-class Java developer with an eagle eye for unintended bugs and edge cases. You carefully explain code with great detail and accuracy. You organize your explanations in markdown-formatted, bulleted lists.
        You write careful, accurate unit tests. When asked to reply only with code, you write all of your code in a single block.
        A good unit test suite should aim to:
        - Test the function's behavior for a wide range of possible inputs
        - Test edge cases that the author may not have foreseen
        - Take advantage of the features of `pytest` to make the tests easy to write and maintain
        - Be easy to read and understand, with clean code and descriptive names
        - Be deterministic, so that the tests always pass or fail in the same way
        Use the following pieces of MemoryContext to answer the question at the end. Also remember ConversationHistory is a list of Conversation objects.
        ---
        ConversationHistory: {history}
        ---
        MemoryContext: {context}
        ---
        Human: {question}
        Bot:"""    
    
    # Retrieve chunks based on the question and assemble them into a joined context:
    """
    Lock the user in a loop to keep asking questions until they type 'exit'
    Add the LLM's answer to the chat history to keep the feedback more conversational
    """
    while True:
        question = input("Ask a question > ")
        if question == 'exit':
            break
        else:
            chunks = search_store(store, question)
            contexts = []
            for i, chunk in enumerate(chunks):
                contexts.append(f"Context {i}:\n{chunk.page_content}")
            if (write_to_disk):
                with open('contexts.json', 'w') as f:
                    json.dump(contexts, f)

            joined_contexts = "\n\n".join(contexts)
            
            prompt = Prompt(template=promptTemplate, input_variables=["context", "question", "history"])
            llmChain = LLMChain(prompt=prompt, llm=ChatOpenAI(model=model,temperature=0))
            # If user's asked to show the context, provide it to them (chunks of text from their vector store):
            if (show_context):
                print(f"Context Provided: \n{joined_contexts}")

            # For each message to OpenAI, print tokens used for each part and in total
            question_tokens = llmChain.llm.get_num_tokens(question)
            contexts_tokens = llmChain.llm.get_num_tokens(joined_contexts)
            history_tokens = llmChain.llm.get_num_tokens(history)
            print("Question tokens: " + str(question_tokens) + "\n" +
                "Contexts' tokens: " + str(contexts_tokens) + "\n" +
                "History tokens: " + str(history_tokens) + "\n\n" +
                "TOTAL: " + str(question_tokens+contexts_tokens+history_tokens))
            
            # Return the prediction.
            answer = llmChain.predict(prompt=prompt,
                            question=question, 
                            context=joined_contexts,
                            history=history)
            history = history + answer +"\n\n###\n\n"
            print(f"Bot: {answer}")
            print("Answer tokens: " + str(llmChain.llm.get_num_tokens(answer)))



from data_chunker import parser
from data_chunker import java_code as JCC
import faiss
from langchain import OpenAI, LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.embeddings import OpenAIEmbeddings
from langchain.prompts import Prompt
from langchain.vectorstores import FAISS
import tiktoken

# Print parsing statistics
def print_parsing_statistics(t: list, f: list):
    attempts = len(t)
    failures = len(f)
    print("\n## Parsing Statistics ##")
    print("Number of files attempted to be parsed = " + str(attempts))
    print("Number of failed files = " + str(failures) +
          ", Failure rate = " + "{:.2f}".format(failures/attempts*100) + "%")

def print_token_statistics(chunks: list, token_limit:int=1600):
    # Print token statistics
    encoding = tiktoken.get_encoding("cl100k_base")
    encoding = tiktoken.encoding_for_model("gpt-3.5-turbo")
    num_member_tokens = num_method_tokens = max_tokens = num_method_chunks = 0
    num_member_chunks = len(chunks)
    max_chunk = None
    max_chunks = []
    for chunk in chunks:
        tokens = len(encoding.encode(str(chunk)))
        num_member_tokens = num_member_tokens + tokens
        if chunk['member'] == "method":
            num_method_tokens = num_method_tokens + tokens
            num_method_chunks = num_method_chunks + 1
        if tokens > max_tokens:
            max_tokens = tokens
            max_chunk = chunk
        if tokens > token_limit:
            max_chunks.append({chunk['typename'], chunk['membername'], tokens})
    print("\n## Token Statistics ##")
    print("Number of chunks generated = " + str(num_member_chunks))
    print("Average number of tokens per chunk = " + 
          "{:.2f}".format(num_member_tokens/num_member_chunks))
    print("Average number of tokens per method chunk = " +
          "{:.2f}".format(num_method_tokens/num_method_chunks))
    print("Maximum token size = " + str(max_tokens))
    print("Number of chunks over " + str(token_limit) + 
          " is " + str(len(max_chunks)))
    #for chunk in max_chunks:
    #    print(chunk)

def print_chunks_sample(chunks: list):
    print("Chunks sample")
    print("-------------")
    for chunk in chunks[:10]:
        print(str(chunk))
    print("...")

def train(chunks:list):
    str_chunks = []
    for chunk in chunks:
        str_chunks.append(str(chunk))

    return FAISS.from_texts(str_chunks, OpenAIEmbeddings())

def chunk(file_path:str, fileExtension:str):
    # Retrieve file list and initialize lists
    data = parser.get_file_list(file_path, fileExtension)
    chunks = []
    failed_files = []
    # Chunk data using the files in the training data
    if fileExtension == "*.java":
        for file in data:
            codelines = parser.get_code_lines(file)
            try:
                tree = JCC.parse_code(file, codelines)
            except JCC.ParseError as e:
                failed_files.append(str(file) + ": " + str(e))
            if tree != None:
                try:
                    chunks = chunks + JCC.chunk_constants(tree)
                    chunks = chunks + JCC.chunk_constructors(tree, codelines)
                    chunks = chunks + JCC.chunk_fields(tree, codelines)
                    chunks = chunks + JCC.chunk_methods(tree, codelines)
                except JCC.ChunkingError as e:
                    failed_files.append(str(file) + ": " + str(e))
            else:
                failed_files.append(str(file) + ", has no tree!")
    else:
        print(f'''File extension type "{fileExtension}" is currently not supported.''')
        exit()

    return data, chunks, failed_files

def on_message(question, history, store, llmchain, show_context:bool=False):
    # Retrieve chunks based on the question and assemble them into a 
    # joined context.
    chunks = store.similarity_search(question)
    contexts = []
    for i, chunk in enumerate(chunks):
        contexts.append(f"Context {i}:\n{chunk.page_content}")
    joined_contexts = "\n\n".join(contexts)
    if show_context:
        print(f"Context Provided:\n {joined_contexts}")
    # For each message to OpenAI, print tokens used for each part and in total
    question_tokens = llmchain.llm.get_num_tokens(question)
    contexts_tokens = llmchain.llm.get_num_tokens(joined_contexts)
    history_tokens = llmchain.llm.get_num_tokens(history)
    print("Question tokens: " + str(question_tokens) + ", " +
          "Contexts' tokens: " + str(contexts_tokens) + ", " +
          "History tokens: " + str(history_tokens) + ", " +
          "TOTAL: " + str(question_tokens+contexts_tokens+history_tokens))
    # Return the prediction.
    return llmchain.predict(question=question, 
                            context=joined_contexts,
                            history=history)

def open_master_prompt(master_prompt_path:str):
    with open(master_prompt_path, "r") as f:
        promptTemplate = f.read()
    return Prompt(template=promptTemplate, input_variables=["context", "question", "history"])

def files_prompter(store, master_prompt_path:str, show_context=False):
    prompt = open_master_prompt(master_prompt_path)
    llmchain = LLMChain(prompt=prompt, llm=ChatOpenAI(model="gpt-3.5-turbo",temperature=0))
    print("attempted file prompt")

def prompter(store, master_prompt_path:str, show_context:bool=False):
    prompt = open_master_prompt(master_prompt_path)
    llmchain = LLMChain(prompt=prompt, llm=ChatOpenAI(model="gpt-3.5-turbo",temperature=0))
    history = ""
    while True:
        question = input("\nAsk a question: ")
        if question == 'exit':
            break
        else:
            answer = on_message(question, history, store, llmchain, show_context)
            history = history + answer +"\n\n###\n\n"
            print(f"\nBot: {answer}")
            print("Answer tokens: " + str(llmchain.llm.get_num_tokens(answer)))